"""Hellinger distance module."""

from typing import List, Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedBins,
)


class HellingerDistance(BaseDistanceBasedBins):
    """Hellinger distance [hellinger1909neue]_ detector.

    :References:

    .. [hellinger1909neue] Hellinger, Ernst.
        "Neue begründung der theorie quadratischer formen von unendlichvielen
        veränderlichen."
        Journal für die reine und angewandte Mathematik 1909.136 (1909): 210-271.
    """

    def __init__(
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param num_bins: number of bins in which to divide probabilities
        :type num_bins: int
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        sqrt_div = np.sqrt(2)
        super().__init__(
            statistical_method=self._hellinger,
            statistical_kwargs={
                "num_bins": num_bins,
                "sqrt_div": sqrt_div,
            },
            callbacks=callbacks,
        )
        self.num_bins = num_bins
        self.sqrt_div = sqrt_div

    def _distance_measure_bins(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
    ) -> float:
        hellinger = self._hellinger(
            X=X_ref,
            Y=X,
            num_bins=self.num_bins,
            sqrt_div=self.sqrt_div,
        )
        return hellinger

    @staticmethod
    def _hellinger(
        X: np.ndarray, Y: np.ndarray, *, num_bins: int, sqrt_div: float  # noqa: N803
    ) -> float:
        (  # noqa: N806
            X_percents,
            Y_percents,
        ) = BaseDistanceBasedBins._calculate_bins_values(
            X_ref=X, X=Y, num_bins=num_bins
        )
        hellinger = (
            np.sqrt(np.sum((np.sqrt(X_percents) - np.sqrt(Y_percents)) ** 2)) / sqrt_div
        )
        return hellinger



"""KSTest (Kolmogorov-Smirnov test) module."""

from typing import Optional, List, Union

import numpy as np  # type: ignore
from scipy.stats import ks_2samp  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class KSTest(BaseStatisticalTest):
    """KSTest (Kolmogorov-Smirnov test) [massey1951kolmogorov]_ detector.

    :References:

    .. [massey1951kolmogorov] Massey Jr, Frank J.
        "The Kolmogorov-Smirnov test for goodness of fit."
        Journal of the American statistical Association 46.253 (1951): 68-78.
    """

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        test = ks_2samp(
            data1=X_ref,
            data2=X,
            alternative=kwargs.get("alternative", "two-sided"),
            method=kwargs.get("method", "auto"),
        )
        test = StatisticalResult(statistic=test.statistic, p_value=test.pvalue)
        return test



"""Anderson-Darling test module."""

from typing import Optional, List, Union

import numpy as np  # type: ignore
from scipy.stats import anderson_ksamp  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class AndersonDarlingTest(BaseStatisticalTest):
    """Anderson-Darling test [scholz1987k]_ detector.

    :Note:
     p-values are bounded between 0.001 and 0.25 according to scipy documentation [1]_.

    :References:

    .. [scholz1987k] Scholz, Fritz W., and Michael A. Stephens.
        "K-sample Anderson–Darling tests."
        Journal of the American Statistical Association 82.399 (1987): 918-924.
       [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html  # noqa: E501 # pylint: disable=line-too-long
    """

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        test = anderson_ksamp(
            samples=[
                X_ref,
                X,
            ],
            **kwargs,
        )
        test = StatisticalResult(
            statistic=test.statistic,
            p_value=test.pvalue,
        )
        return test



"""Mann-Whitney U test module."""

from typing import Optional, List, Union

import numpy as np  # type: ignore
from scipy.stats import mannwhitneyu  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class MannWhitneyUTest(BaseStatisticalTest):
    """Mann–Whitney U test [mann1947test]_ detector.

    :References:

    .. [mann1947test] Mann, Henry B., and Donald R. Whitney.
        "On a test of whether one of two random variables is stochastically larger than
        the other."
        The annals of mathematical statistics (1947): 50-60.
    """

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        test = mannwhitneyu(  # pylint: disable=unexpected-keyword-arg
            x=X_ref,
            y=X,
            alternative="two-sided",
            nan_policy="raise",
            **kwargs,
        )
        test = StatisticalResult(
            statistic=test.statistic,
            p_value=test.pvalue,
        )
        return test



"""CVMTest (Cramér-von Mises test) module."""

from typing import Optional, List, Union

import numpy as np  # type: ignore
from scipy.stats import cramervonmises_2samp  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)
from frouros.detectors.data_drift.exceptions import InsufficientSamplesError


class CVMTest(BaseStatisticalTest):
    """CVMTest (Cramér-von Mises test) [cramer1928composition]_ detector.

    :References:

    .. [cramer1928composition] Cramér, Harald.
        "On the composition of elementary errors: First paper: Mathematical deductions."
        Scandinavian Actuarial Journal 1928.1 (1928): 13-74.
    """

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    @BaseStatisticalTest.X_ref.setter  # type: ignore[attr-defined]
    def X_ref(self, value: Optional[np.ndarray]) -> None:  # noqa: N802
        """Reference data setter.

        :param value: value to be set
        :type value: Optional[numpy.ndarray]
        """
        if value is not None:
            self._check_sufficient_samples(X=value)
            self._X_ref = value
            # self._X_ref_ = check_array(value)  # noqa: N806
        else:
            self._X_ref = None  # noqa: N806

    def _specific_checks(self, X: np.ndarray) -> None:  # noqa: N803
        self._check_sufficient_samples(X=X)

    @staticmethod
    def _check_sufficient_samples(X: np.ndarray) -> None:  # noqa: N803
        if X.shape[0] < 2:
            raise InsufficientSamplesError("Number of samples must be at least 2.")

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        test = cramervonmises_2samp(
            x=X_ref,
            y=X,
            method=kwargs.get("method", "auto"),
        )
        test = StatisticalResult(statistic=test.statistic, p_value=test.pvalue)
        return test



"""EMD (Earth Mover's Distance) module."""

from typing import List, Optional, Union

import numpy as np  # type: ignore
from scipy.stats import wasserstein_distance  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import UnivariateData
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)


class EMD(BaseDistanceBased):
    """EMD (Earth Mover's Distance) [rubner2000earth]_ detector.

    :References:

    .. [rubner2000earth] Rubner, Yossi, Carlo Tomasi, and Leonidas J. Guibas.
        "The earth mover's distance as a metric for image retrieval."
        International journal of computer vision 40.2 (2000): 99.
    """

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]] = None,
        **kwargs,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, List[BaseCallbackBatch]]]
        """
        super().__init__(
            statistical_type=UnivariateData(),
            statistical_method=self._emd,
            statistical_kwargs=kwargs,
            callbacks=callbacks,
        )
        self.kwargs = kwargs

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        emd = self._emd(X=X_ref, Y=X, **self.kwargs)
        distance = DistanceResult(distance=emd)
        return distance

    @staticmethod
    def _emd(X: np.ndarray, Y: np.ndarray, **kwargs) -> float:  # noqa: N803
        emd = wasserstein_distance(
            u_values=X.flatten(),
            v_values=Y.flatten(),
            **kwargs,
        )
        return emd



"""Reset batch callback module."""

from typing import Optional

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.utils.logger import logger


class ResetStatisticalTest(BaseCallbackBatch):
    """Reset on statistical test batch callback class."""

    def __init__(self, alpha: float, name: Optional[str] = None) -> None:
        """Init method.

        :param alpha: significance value
        :type alpha: float
        :param name: name to be use
        :type name: Optional[str]
        """
        super().__init__(name=name)
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Alpha property.

        :return: significance value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Alpha setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("value must be greater than 0.")
        self._alpha = value

    def on_compare_end(self, **kwargs) -> None:
        """On compare end method."""
        p_value = kwargs["result"].p_value
        if p_value < self.alpha:
            logger.info("Drift detected. Resetting detector...")
            self.detector.reset()  # type: ignore

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseDataDriftBatch):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseDataDriftBatch."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""



"""mSPRT (Mixing Sequentially Probability Ratio Test) callback module."""

from typing import Union, Tuple, Optional

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.utils.stats import CircularMean


class mSPRT(BaseCallbackStreaming):  # noqa: N801 # pylint: disable=invalid-name
    """mSPRT (mixing Sequentially Probability Ratio Test) callback class.

    :References:

    .. [johari2022always] Ramesh, Johari, et al.
        "Always valid inference: Continuous monitoring of a/b tests"
        Operations Research 70.3 (2022): 1806-1821.
    """

    def __init__(
        self,
        alpha: float,
        sigma: Union[int, float] = 1.0,
        tau: Union[int, float] = 1.0,
        lambda_: Union[int, float] = 1.0,
        name: Optional[str] = None,
    ) -> None:
        """Init method.

        :param alpha: alpha value
        :type alpha: float
        :param sigma: sigma value
        :type sigma: Union[int, float]
        :param tau: tau value
        :type tau: Union[int, float]
        :param lambda_: lambda value
        :type lambda_: Union[int, float]
        :param name: name value
        :type name: Optional[str]
        """
        super().__init__(name=name)
        self.alpha = alpha
        self.sigma = sigma
        self.sigma_squared = self.sigma**2
        self.two_sigma_squared = 2 * self.sigma_squared
        self.tau = tau
        self.tau_squared = self.tau**2
        self.lambda_ = lambda_
        self.mean = None
        self.theta = None
        self.p_value = 1.0

    @property
    def alpha(self) -> float:
        """Alpha property.

        :return: alpha value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Alpha setter.

        :param value: value to be set
        :type value: float
        """
        if not isinstance(value, float):
            raise TypeError("alpha must be a float")
        if not 0.0 < value < 1.0:
            raise ValueError("alpha must be in the range (0, 1)")
        self._alpha = value

    @property
    def sigma(self) -> Optional[Union[int, float]]:
        """Sigma property.

        :return: sigma value
        :rtype: Optional[Union[int, float]]
        """
        return self._sigma

    @sigma.setter
    def sigma(self, value: Optional[Union[int, float]]) -> None:
        """Sigma setter.

        :param value: value to be set
        :type value: Optional[float]
        """
        if value is not None and not isinstance(value, (int, float)):
            raise TypeError("sigma must be int, float or None")
        self._sigma = value

    @property
    def tau(self) -> Optional[Union[int, float]]:
        """Tau property.

        :return: tau squared value
        :rtype: Optional[Union[int, float]]
        """
        return self._tau

    @tau.setter
    def tau(self, value: Union[int, float]) -> None:
        """Tau setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        if not isinstance(value, (int, float)):
            raise TypeError("tau must be int, float or None")
        self._tau = value

    @property
    def lambda_(self) -> Optional[Union[int, float]]:
        """Lambda property.

        :return: lambda value
        :rtype: Optional[Union[int, float]]
        """
        return self._lambda_

    @lambda_.setter
    def lambda_(self, value: Union[int, float]) -> None:
        """Lambda setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        if not isinstance(value, (int, float)):
            if value <= 0.0:
                raise ValueError("lambda_ must be greater than 0")
        self._lambda_ = value

    def on_fit_end(self, **kwargs) -> None:
        """On fit end method."""
        self.mean = CircularMean(size=self.detector.window_size)  # type: ignore
        self.theta = self.detector.compare(X=kwargs["X"])[0].distance  # type: ignore

    def on_update_end(self, **kwargs) -> None:
        """On update end method."""
        self.mean.update(value=kwargs["value"])  # type: ignore
        self.p_value, likelihood = self._calculate_p_value()

        self.logs.update(
            {
                "distance_mean": self.mean.get(),  # type: ignore
                "likelihood": likelihood,
                "p_value": self.p_value,
            },
        )

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.mean = None
        self.p_value = 1.0

    def _calculate_p_value(self) -> Tuple[float, float]:
        likelihood = self._likelihood_normal_mixing_distribution(
            mean=self.mean.get(),  # type: ignore
            sigma=self.sigma,  # type: ignore
            sigma_squared=self.sigma_squared,
            tau_squared=self.tau_squared,
            two_sigma_squared=self.two_sigma_squared,
            n=self.detector.window_size,  # type: ignore
            theta=self.theta,  # type: ignore
            lambda_=self.lambda_,  # type: ignore
        )
        p_value = min(
            self.p_value,
            1 / likelihood,
        )
        return p_value, likelihood

    @staticmethod
    def _likelihood_normal_mixing_distribution(  # pylint: disable=too-many-arguments
        mean: float,
        sigma: float,
        sigma_squared: float,
        tau_squared: float,
        two_sigma_squared: float,
        n: int,
        theta: float,
        lambda_: float,
    ) -> float:
        # FIXME: Explore lambda_ influence   # pylint: disable=fixme
        #  and redesign the likelihood formula
        n_tau_squared = n * tau_squared
        sigma_squared_plus_n_tau_squared = sigma_squared + n_tau_squared
        likelihood = (sigma / np.sqrt(sigma_squared_plus_n_tau_squared)) * np.exp(
            n
            * n_tau_squared
            * lambda_  # Not present in mSPRT, added as a hyperparameter to control
            # the influence of the distance difference
            * (mean - theta)
            ** 2  # (mean-theta) ** 2, theta=detector statistic (H_0 value, no distance)
            / (two_sigma_squared * sigma_squared_plus_n_tau_squared)
        )
        return likelihood



"""Geometric Moving Average module."""

from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseCUSUM,
    BaseCUSUMConfig,
    AlphaConfig,
)


class GeometricMovingAverageConfig(BaseCUSUMConfig, AlphaConfig):
    """Geometric Moving Average [robertst1959control]_ configuration.

    :References:

    .. [robertst1959control] Roberts, S. W.
        “Control Chart Tests Based on Geometric Moving Averages.”
        Technometrics, vol. 1, no. 3, 1959, pp. 239–50.
        JSTOR, https://doi.org/10.2307/1266443.
    """

    def __init__(
        self,
        alpha: float = 0.99,
        lambda_: float = 1.0,
        min_num_instances: int = 30,
    ) -> None:
        """Init method.

        :param alpha: forgetting factor value
        :type alpha: float
        :param lambda_: delta value
        :type lambda_: float
        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        """
        BaseCUSUMConfig.__init__(
            self, lambda_=lambda_, min_num_instances=min_num_instances
        )
        AlphaConfig.__init__(self, alpha=alpha)


class GeometricMovingAverage(BaseCUSUM):
    """Geometric Moving Average [robertst1959control]_ detector.

    :References:

    .. [robertst1959control] Roberts, S. W.
        “Control Chart Tests Based on Geometric Moving Averages.”
        Technometrics, vol. 1, no. 3, 1959, pp. 239–50.
        JSTOR, https://doi.org/10.2307/1266443.
    """

    config_type = GeometricMovingAverageConfig  # type: ignore

    def _update_sum(self, error_rate: float) -> None:
        self.sum_ = self.config.alpha * self.sum_ + (  # type: ignore
            1 - self.config.alpha  # type: ignore
        ) * (error_rate - self.mean_error_rate.mean)



"""Permutation test batch callback module."""

import multiprocessing
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np  # type: ignore
from scipy.stats import norm  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.utils.stats import permutation, z_score


class PermutationTestDistanceBased(BaseCallbackBatch):
    """Permutation test on distance based batch callback class."""

    def __init__(
        self,
        num_permutations: int,
        num_jobs: int = -1,
        name: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """Init method.

        :param num_permutations: number of permutations
        :type num_permutations: int
        :param num_jobs: number of jobs
        :type num_jobs: int
        :param name: name to be use
        :type name: Optional[str]
        """
        super().__init__(name=name)
        self.num_permutations = num_permutations
        self.num_jobs = num_jobs
        self.verbose = verbose
        self.permutation_kwargs = kwargs

    @property
    def num_permutations(self) -> int:
        """Number of permutations property.

        :return: number of permutation to obtain the p-value
        :rtype: int
        """
        return self._num_permutations

    @num_permutations.setter
    def num_permutations(self, value: int) -> None:
        """Number of permutations method setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("value must be greater of equal than 1.")
        self._num_permutations = value

    @property
    def num_jobs(self) -> int:
        """Number of jobs property.

        :return: number of jobs to use
        :rtype: int
        """
        return self._num_jobs

    @num_jobs.setter
    def num_jobs(self, value: int) -> None:
        """Number of jobs method setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value == 0 or value < -1:
            raise ValueError("value must be greater than 0 or -1.")
        self._num_jobs = multiprocessing.cpu_count() if value == -1 else value

    @property
    def verbose(self) -> bool:
        """Verbose flag property.

        :return: verbose flag
        :rtype: bool
        """
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """Verbose flag setter.

        :param value: value to be set
        :type value: bool
        :raises TypeError: Type error exception
        """
        if not isinstance(value, bool):
            raise TypeError("value must of type bool.")
        self._verbose = value

    @staticmethod
    def _calculate_p_value(  # pylint: disable=too-many-arguments
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
        statistic: Callable,
        statistic_args: Dict[str, Any],
        observed_statistic: float,
        num_permutations: int,
        num_jobs: int,
        random_state: int,
        verbose: bool,
    ) -> Tuple[List[float], float]:
        permuted_statistic = permutation(
            X=X_ref,
            Y=X_test,
            statistic=statistic,
            statistical_args=statistic_args,
            num_permutations=num_permutations,
            num_jobs=num_jobs,
            random_state=random_state,
            verbose=verbose,
        )
        permuted_statistic = np.array(permuted_statistic)
        # Use z-score to calculate p-value
        observed_z_score = z_score(
            value=observed_statistic,
            mean=permuted_statistic.mean(),  # type: ignore
            std=permuted_statistic.std(),  # type: ignore
        )
        p_value = norm.sf(np.abs(observed_z_score)) * 2
        return permuted_statistic, p_value

    def on_compare_end(self, **kwargs) -> None:
        """On compare end method."""
        X_ref, X_test = kwargs["X_ref"], kwargs["X_test"]  # noqa: N806
        observed_statistic = kwargs["result"][0]
        permuted_statistics, p_value = self._calculate_p_value(
            X_ref=X_ref,
            X_test=X_test,
            statistic=self.detector.statistical_method,  # type: ignore
            statistic_args=self.detector.statistical_kwargs,  # type: ignore
            observed_statistic=observed_statistic,
            num_permutations=self.num_permutations,
            num_jobs=self.num_jobs,
            verbose=self.verbose,
            **self.permutation_kwargs,
        )
        self.logs.update(
            {
                "observed_statistic": observed_statistic,
                "permuted_statistics": permuted_statistics,
                "p_value": p_value,
            },
        )

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseDataDriftBatch):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseDataDriftBatch."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""



"""Page Hinkley module."""

from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseCUSUM,
    BaseCUSUMConfig,
    DeltaConfig,
    AlphaConfig,
)


class PageHinkleyConfig(BaseCUSUMConfig, DeltaConfig, AlphaConfig):
    """Page Hinkley [page1954continuous]_ configuration.

    :References:

    .. [page1954continuous] Page, Ewan S.
        "Continuous inspection schemes."
        Biometrika 41.1/2 (1954): 100-115.
    """

    def __init__(
        self,
        delta: float = 0.005,
        lambda_: float = 50.0,
        min_num_instances: int = 30,
        alpha: float = 0.9999,
    ) -> None:
        """Init method.

        :param delta: delta value
        :type delta: float
        :param lambda_: lambda value
        :type lambda_: float
        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        :param alpha: forgetting factor value
        :type alpha: float
        """
        BaseCUSUMConfig.__init__(
            self, min_num_instances=min_num_instances, lambda_=lambda_
        )
        DeltaConfig.__init__(self, delta=delta)
        AlphaConfig.__init__(self, alpha=alpha)


class PageHinkley(BaseCUSUM):
    """Page Hinkley [page1954continuous]_ detector.

    :References:

    .. [page1954continuous] Page, Ewan S.
        "Continuous inspection schemes."
        Biometrika 41.1/2 (1954): 100-115.
    """

    config_type = PageHinkleyConfig  # type: ignore

    def _update_sum(self, error_rate: float) -> None:
        self.sum_ = self.config.alpha * self.sum_ + (  # type: ignore
            error_rate - self.mean_error_rate.mean - self.config.delta  # type: ignore
        )



"""Real datasets module."""

from typing import Optional

import numpy as np  # type: ignore
from scipy.io import arff  # type: ignore

from frouros.datasets.base import BaseDatasetDownload


class Elec2(BaseDatasetDownload):
    """Elec2 dataset [harries1999splice]_.

    :References:

    .. [harries1999splice] Harries, Michael.
        "Splice-2 comparative evaluation: Electricity pricing." (1999).
    """

    def __init__(self, file_path: Optional[str] = None) -> None:
        """Init method.

        :param file_path: file path for the downloaded file
        :type file_path: Optional[str]
        """
        super().__init__(
            url=[
                "https://nextcloud.ifca.es/index.php/s/2coqgBEpa82boLS/download",
                "https://www.openml.org/data/download/2419/electricity-normalized.arff",
            ],
            file_path=file_path,
        )

    def read_file(self, **kwargs) -> np.ndarray:
        """Read file.

        :param kwargs: dict of kwargs
        :type kwargs: dict
        :return: read file
        :rtype: numpy.ndarray
        """
        index = kwargs.get("index", 0)
        dataset = arff.loadarff(f=self.file_path)[index]
        return dataset



"""Warning samples buffer callback module."""

import copy
from typing import Any, List, Optional

from frouros.callbacks.streaming.base import BaseCallbackStreaming


class WarningSamplesBuffer(BaseCallbackStreaming):
    """Store warning samples as a buffer callback class."""

    def __init__(self, name: Optional[str] = None) -> None:
        """Init method.

        :param name: name to be use
        :type name: Optional[str]
        """
        super().__init__(name=name)
        self.X: List[Any] = []
        self.y: List[Any] = []
        self._start_warning = False

    def on_update_start(self, **kwargs) -> None:
        """On update start method."""
        self._start_warning = not self.detector.warning  # type: ignore

    def on_update_end(self, **kwargs) -> None:
        """On update end method."""
        self.logs = {
            "X": copy.deepcopy(self.X),
            "y": copy.deepcopy(self.y),
        }

    def on_warning_detected(self, **kwargs) -> None:
        """On warning detected method."""
        if self._start_warning:
            map(lambda x: x.clear(), [self.X, self.y])
        self.X.append(kwargs["X"])
        self.y.append(kwargs["y"])

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, DDMBased):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type DDMBased."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""
        self.X.clear()
        self.y.clear()
        self._start_warning = False
